#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <png.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <vga.h>
#include <vgagl.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>

#include "config.h"

/* maximums */
#define MAXSTARS (300)		/* number of stars */

#define MAXSCREENSHOTWIDTH (512)
#define MAXSCREENSHOTHEIGHT (512)

#define MAXGAMES 3000		/* max number of roms */
#define MAXDESC  23		/* max length of rom description */

#define MUSICMAX 3		/* number of mp3s */
#define NUMLETTERS 44

/* graphics */
#define GRAPHICSMODE  G320x240x64K
#define TEXTSIZE 9.3			/* font size */
#define TEXTSIZEINT 10			/* font size, in integer form (round up) */

typedef struct {			/* single graphical letter */
	int ascii;
	png_bytep row[TEXTSIZEINT];
} letterT;

letterT letter[NUMLETTERS];		/* all letters */

typedef struct {			/* rgb tuple */
	int r;
	int g;
	int b;
} colorT;

colorT textcolor = {140, 140, 0};
colorT starcolor[6];
colorT cursorcolor[] = {{255, 0, 0},{0, 255, 0}};


int brighten = 255;

/* control keys (ascii codes */
#define DOWN  		(115)
#define UP    		(119)

#define HOME 6
#define END 104

#define QUIT 		(113)

#define PGUP 		(91)
#define PGDOWN 		(93)

#define SLASH 		(47)
#define QUESTION 	(63)
#define NEXTSEARCH 	(110) 

#define PNGWRITE (112)

#define ENTER (10)


/* coords */
#define XS 7			/* used to position cursor */
#define YS 7
#define SCREENSHOTX (160)	/* base coordinates for screenshots*/
#define SCREENSHOTY (112)
#define TITLEINDENT   1		/* number of blanks before rom description */
#define YEARINDENT    27	/* number of blanks before rom year */
#define IMAGEDELAY 15		/* delay before moving cursor and 
				   screenshot being loaded */
#define XGRID 10		/* positioning of text */
#define YGRID 10


int shotxoff = 0;	/* how far screenshot has moved from base */
int shotyoff = 0;
int oldshotxoff = -99;	/* for determining whether screenshot has moved */
int oldshotyoff = -99;
int shotxspeed = 12;	/* how fast will the screenshot slide on/off? */
int shotyspeed = 8;
int imagedir = -1;	/* -1 = screenshot is sliding onto the screen 
			    0 = screenshot is off the screen, new image to 
			    	be loaded
			   +1 = screenshot is sliding off the scree */

/* random number routines */
#define randomize() srand((unsigned)time(NULL))
#define random(n) (rand() % n)

/* screens */
GraphicsContext *physicalscreen;
GraphicsContext *virtualscreen;
GraphicsContext *virtualscreen2;


/* processes */
long musicpid = 0; /* pid of mpg123 */

long nlines;	/* number of lines of rom descriptions to draw*/
int err;	/* is an error message going to be displayed? */

/* misc */
int mod[61][6];	/* modulus lookup table, we actually only use 1-60, 1-5 */
int count = 1;	/* iteration counter, used for timing */
int nothingcount = IMAGEDELAY;	/* how long has the cursor been still for? */
int needupdate = 0;	/* do we need to redraw the screen? */
int newpicture = 0;	/* do we need to load a new screenshot? */

/* cursor */
long curoff = 0; /* current position of top of screen in rom list*/
int currentcursor = 0;

/* starfield */
typedef struct {
	int color;
	int x;
	int y;
	int speed;
} starT;

starT star[MAXSTARS];

/* roms */
typedef struct {
	char title[10];
	char year[5];
	char desc[80];
} gameT;

gameT game[MAXGAMES];


/* screenshots */
png_bytep row_pointers[MAXSCREENSHOTHEIGHT]; 	/* current screenshot */
png_uint_32 width, height;			/* ... of current screenshot */


/* decs */
void mvaddstr(int y, int x, char *txt);
void getpixelrgb(int x, int y, int *r, int *g, int *b);
void setpixelrgb(int x, int y, int r, int g, int b);
void linergb(int x1,int y1,int x2,int y2, int r, int g, int b);

/* graphics wrapper functions */
void virttophys(void) {
	gl_copyscreen(physicalscreen);
}

void getpixelrgb(int x, int y, int *r, int *g, int *b) {
	gl_getpixelrgb(x,y,r,g,b);
}

void setpixelrgb(int x, int y, int r, int g, int b) {
	gl_setpixelrgb(x,y,r,g,b);
}

void linergb(int x1,int y1,int x2,int y2, int r, int g, int b) {
  int i, deltax, deltay, numpixels, d, dinc1, dinc2, x, xinc1, xinc2, y, yinc1, yinc2;

  // Calculate deltax and deltay for initialisation 
  deltax = abs(x2 - x1);
  deltay = abs(y2 - y1);

  // Initialize all vars based on which is the independent variable 
  if (deltax >= deltay) { 

      // x is independent variable 
      numpixels = deltax + 1;
      d = (2 * deltay) - deltax;
      dinc1 = deltay << 1;
      dinc2 = (deltay - deltax) << 1;
      xinc1 = 1;
      xinc2 = 1;
      yinc1 = 0;
      yinc2 = 1;
  } else {
      // y is independent variable 
      numpixels = deltay + 1;
      d = (2 * deltax) - deltay;
      dinc1 = deltax << 1;
      dinc2 = (deltax - deltay) << 1;
      xinc1 = 0;
      xinc2 = 1;
      yinc1 = 1;
      yinc2 = 1;
  }

  // Make sure x and y move in the right directions 
  if (x1 > x2) {
      xinc1 = - xinc1;
      xinc2 = - xinc2;
  }
  if (y1 > y2) {
      yinc1 = - yinc1;
      yinc2 = - yinc2;
  }
  // Start drawing at <x1, y1> 
  x = x1;
  y = y1;

  // Draw the pixels 
  for (i = 0; i < numpixels; i++) {
      setpixelrgb(x, y, r,g,b);
      if (d < 0) {
          d = d + dinc1;
          x = x + xinc1;
          y = y + yinc1;
      } else {
          d = d + dinc2;
          x = x + xinc2;
          y = y + yinc2;
      }
  }

}

/* display currently loaded screenshot */
void show_png(void) {

	int y,xx,yy,count;
	int startx,starty;
	png_bytep x;
   /* draw image */

   /* centre it */
   startx = SCREENSHOTX + ((320-SCREENSHOTX) - (width/2))/2;
   starty = SCREENSHOTY + ((240-SCREENSHOTY) - (height/2))/2;
   
   /* clear it first */
   /* background portion */
   
   for (yy = SCREENSHOTY; yy < SCREENSHOTY+shotyoff; yy++) {
	   linergb(SCREENSHOTX,yy,319,yy,0,0,0);
   }
   for (yy = SCREENSHOTY+shotyoff; yy <= 239; yy++) {
	   linergb(SCREENSHOTX,yy,SCREENSHOTX+shotxoff-1,yy,0,0,0);
   }
   
   /* screenshot portion */
   for (yy = SCREENSHOTY+shotyoff; yy <= 239; yy++) {
	   linergb(SCREENSHOTX+shotxoff,yy,319,yy,0,0,0);
   }
   
   
   xx = startx + shotxoff;
   yy = starty + shotyoff;

   for (y = 0; y < height;y ++ ) {
       count = 0;
       for (x = row_pointers[y]; count < width; ) {
	        int r,g,b,rr,gg,bb;
                //memcpy(&clr2,row_pointers[y]+x*sizeof(int),sizeof(int));
	//	if (row_pointers[y] == NULL) return;
		r = *x; x++;
		g = *x; x++;
		b = *x; x++;
		/* skip 1 pixel */
		x += 3;
		//r = g = b = 255;
		//clr2 = *x;
                rr = (unsigned char) r;
                gg = (unsigned char) g;
                bb = (unsigned char) b;
                //if (clr > 255) clr = 255;
		//if (clr == 0) clr = 255;
		if ((xx <= 319) && (xx >= 0) && (yy <= 239) && (yy >= 0)) {
                	setpixelrgb(xx,yy,rr,gg,bb);
		}
                xx++;
                if (xx > 320) xx = 320;
		count += 2 ;
       }
       yy++;
       xx = startx + shotxoff;
       if (yy > 240) yy = 240;
       /* skip one line */
       y++;
   }

   /* border */
   /*
   linergb(SCREENSHOTX+shotxoff,SCREENSHOTY+shotyoff,319,SCREENSHOTY+shotyoff,textcolor.r,textcolor.g,textcolor.b);
   linergb(SCREENSHOTX+shotxoff,SCREENSHOTY+shotyoff,SCREENSHOTX+shotxoff,239,textcolor.r,textcolor.g,textcolor.b);
   linergb(SCREENSHOTX+shotxoff,239,319,239,textcolor.r,textcolor.g,textcolor.b);
   linergb(319,SCREENSHOTY+shotyoff,319,239,textcolor.r,textcolor.g,textcolor.b);
*/
}


/* load a screenshot into memory (most of this copied straight 
 * out the libpng example code */
int read_png(char *file_name) 
{
   png_structp png_ptr;
   png_infop info_ptr;
   int bit_depth, color_type, interlace_type;
   FILE *fp;
   FILE *f2;
   
   if ((fp = fopen(file_name, "rb")) == NULL) {
	   int tries = 0;
	   /* keep stripping chars off end of name and see if it works */
	   for (;(tries < 3) && (strlen(file_name) >= 2);tries++, file_name[strlen(file_name)-1] = '\0') {
		if (fp = fopen(file_name, "rb")) {
			break;
		}
	   }
	   /* if we still have no success... */
	   if (fp == NULL) {
	   	strcpy(file_name, "nopic.png");
   	   	if ((fp = fopen(file_name, "rb")) == NULL) {
      			return -1;
	   	}
	   }
   }
   
   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also supply the
    * the compiler header file version, so that we know if the application
    * was compiled with a compatible version of the library.  REQUIRED
    */
   png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
      NULL, NULL, NULL);

   if (png_ptr == NULL)
   {
      fclose(fp);
      return -1;
   }

   /* Allocate/initialize the memory for image information.  REQUIRED. */
   info_ptr = png_create_info_struct(png_ptr);
   if (info_ptr == NULL)
   {
      fclose(fp);
      png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
      return -1;
   }

   /* Set error handling if you are using the setjmp/longjmp method (this is
    * the normal method of doing things with libpng).  REQUIRED unless you
    * set up your own error handlers in the png_create_read_struct() earlier.
    */
   if (setjmp(png_ptr->jmpbuf)) {
      /* Free all of the memory associated with the png_ptr and info_ptr */
      png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
      fclose(fp);
      /* If we get here, we had a problem reading the file */
      return -1;
   }

   /* One of the following I/O initialization methods is REQUIRED */
   /* Set up the input control if you are using standard C streams */
   png_init_io(png_ptr, fp);

   /* If we have already read some of the signature */
   //png_set_sig_bytes(png_ptr, sig_read);

   /* The call to png_read_info() gives us all of the information from the
    * PNG file before the first IDAT (image data chunk).  REQUIRED
    */
   png_read_info(png_ptr, info_ptr);

   png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
       &interlace_type, NULL, NULL);

   //f2 = fopen("blah","wt");
   //fprintf(f2, "%d x %d bits is %d type is %d\n",width,height, bit_depth, color_type);
   //fclose(f2);


   /* Expand paletted colors into true RGB triplets */
   if (color_type == PNG_COLOR_TYPE_PALETTE) {
      png_set_expand(png_ptr);
   }

   /* Optional call to gamma correct and add the background to the palette
    * and update info structure.  REQUIRED if you are expecting libpng to
    * update the palette for you (ie you selected such a transform above).
    */
   png_read_update_info(png_ptr, info_ptr);

   /* Allocate the memory to hold the image using the fields of info_ptr. */
   
   /* The easiest way to read the image: */
   /* Now it's time to read the image.  One of these methods is REQUIRED */
   png_read_image(png_ptr, row_pointers);

   /* read rest of file, and get additional chunks in info_ptr - REQUIRED */
   png_read_end(png_ptr, info_ptr);

   /* clean up after the read, and free any memory allocated - REQUIRED */
   png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

   /* close the file */
   fclose(fp);

   /* that's it */
   return 0;
}


/* write a screenshot of the current screen */
void write_png(char *file_name)
{
   int x,y;
   FILE *fp;
   png_structp png_ptr;
   png_infop info_ptr;

   png_bytep row_pointers[240];
   png_bytep p;

   /* open the file */
   fp = fopen(file_name, "wb");
   if (fp == NULL)
      return;

   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also check that
    * the library version is compatible with the one used at compile time,
    * in case we are using dynamically linked libraries.  REQUIRED.
    */
   png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
      			NULL, NULL, NULL);

   if (png_ptr == NULL)
   {
      fclose(fp);
      return;
   }


   /* Allocate/initialize the image information data.  REQUIRED */
   info_ptr = png_create_info_struct(png_ptr);
   if (info_ptr == NULL)
   {
      fclose(fp);
      png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
      return;
   }

   /* Set error handling.  REQUIRED if you aren't supplying your own
    * error hadnling functions in the png_create_write_struct() call.
    */
   if (setjmp(png_ptr->jmpbuf))
   {
      /* If we get here, we had a problem reading the file */
      fclose(fp);
      png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
      return;
   }

   /* One of the following I/O initialization functions is REQUIRED */
   /* set up the output control if you are using standard C streams */
   png_init_io(png_ptr, fp);


   /* Set the image information here.  Width and height are up to 2^31,
    * bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
    * the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
    * PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
    * or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
    * PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
    * currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
    */
   png_set_IHDR(png_ptr, info_ptr, 320, 240, 8, PNG_COLOR_TYPE_RGB,
      PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

   /* Optional gamma chunk is strongly suggested if you have any guess
    * as to the correct gamma of the image.
    */
   png_set_gAMA(png_ptr, info_ptr, 100);

   /* other optional chunks like cHRM, bKGD, tRNS, tIME, oFFs, pHYs, */
   /* note that if sRGB is present the cHRM chunk must be ignored
    * on read and must be written in accordance with the sRGB profile */

   /* Write the file header information.  REQUIRED */
   png_write_info(png_ptr, info_ptr);

   /* Once we write out the header, the compression type on the text
    * chunks gets changed to PNG_TEXT_COMPRESSION_NONE_WR or
    * PNG_TEXT_COMPRESSION_zTXt_WR, so it doesn't get written out again
    * at the end.
    */

   /* The easiest way to write the image (you may have a different memory
    * layout, however, so choose what fits your needs best).  You need to
    * use the first method if you aren't handling interlacing yourself.
    */

   for (y = 0; y < 240; y++) {
	   row_pointers[y] = malloc(320*3);
   }

   for (y = 0; y < 240; y++) {
     p = row_pointers[y];
     for (x = 0; x < 320; x++) {
	     int r,g,b;
	     getpixelrgb(x,y,&r,&g,&b);
	     *p = r; p++;
	     *p = g; p++;
	     *p = b; p++;
     }
   }

   /* One of the following output methods is REQUIRED */
   png_write_image(png_ptr, row_pointers);

   /* You can write optional chunks like tEXt, zTXt, and tIME at the end
    * as well.
    */

   /* It is REQUIRED to call this to finish writing the rest of the file */
   png_write_end(png_ptr, info_ptr);

   /* if you allocated any text comments, free them here */

   /* close the file */
   fclose(fp);

   /* clean up after the write, and free any memory allocated */
   png_destroy_write_struct(&png_ptr, (png_infopp)NULL);

   /* that's it */
   return;
}

/* set up the star arrays */
void initstars(void) {
	int i;
	int j;
	int cutoff = MAXSTARS / 5;
	int curcolor = 0;
	int curspeed = 1; /* from 1 to 5 */
	for (i = 0; i < MAXSTARS; i++) {
		star[i].x = random(320);
		star[i].y = random(240);
		star[i].speed = curspeed;
		star[i].color = curcolor;
		if (((i+1) % cutoff) == 0) {
			/* get darker and slower (further away) */
			curspeed *= 2;
			curcolor++;
		}
	}
	
	/* mod table */
	for (i = 1; i <= 60; i++) {
		for (j = 1; j <= 5; j++) {
			mod[i][j] = i % j;
		}
	}

}

/* make the cursor change colour (goes between two values) */
void swapcursorcolors(void) {
	if (mod[count][4] == 0) {
		currentcursor = 1 - currentcursor;
	}
}



/* can we draw/remove a star at this position, with this colour? */
int isdrawable (int x, int y, int r, int g, int b) {
	int i;
	int startx,starty;
//	FILE *f;
   	startx = SCREENSHOTX + ((320-SCREENSHOTX) - (width/2))/2 + shotxoff;
   	starty = SCREENSHOTY + ((240-SCREENSHOTY) - (height/2))/2 + shotyoff;

	//if ((x >= (SCREENSHOTX+shotxoff)) && (x <= 320) && 
	if ((x >= (startx)) && (x <= (startx+width/2)) && 
		//(y >= (SCREENSHOTY+shotyoff)) && (y <= 240)) {
		(y >= (starty)) && (y <= starty+height/2)) {
		return 0;
	}
	if ((r == 0) && (g == 0) && (b == 0)) {
		return -1;
	}
	for (i = 0; i < 6; i++) {
		if (r == starcolor[i].r) {
			if (g == starcolor[i].g) {
				if (b == starcolor[i].b) {
					return -1;
				}
			}
		}
	}
	//f = fopen("blah","at");
	//fprintf(f, "cant draw on: %d,%d,%d\n",r,g,b);
	//fclose(f);
	
	return 0;
}

/* erase all stars */
void removestars(void) {
	int i;
	int r,g,b;
	for (i = 0; i < MAXSTARS; i++) {
		getpixelrgb(star[i].x, star[i].y,&r,&g,&b);
		if (isdrawable(star[i].x, star[i].y, r,g,b)) {
			setpixelrgb(star[i].x, star[i].y, 0,0,0);
		}
	}
}

/* animate stars (move downwards) */
void movestars(void) {
	int i;

	count++;
	if (count >= 61) {
		count = 1;
	}

	for (i = 0; i < MAXSTARS; i++) {
		//if ((count % star[i].speed) == 0) {
		if (mod[count][star[i].speed] == 0) {
			star[i].y++;
			if (star[i].y > 239) {
				star[i].y = 0;
				star[i].x = random(320);
			}
		}
	}
}


/* draw all stars */
void drawstars(void) {
	int i;
	int r,g,b;
	for (i = 0; i < MAXSTARS; i++) {
		getpixelrgb(star[i].x, star[i].y,&r,&g,&b);
		if (isdrawable(star[i].x, star[i].y, r,g,b)) {
			setpixelrgb(star[i].x, star[i].y, starcolor[star[i].color].r,
							starcolor[star[i].color].g,
							starcolor[star[i].color].b);
		}
	}
}


/* draw a single character on the screen
 * c = char to draw
 * cx = x position
 * cy = y position 
 * clr = colour
 */
int drawchar(char c,int cy, int cx) {
	int x,y;
	int xx,yy;
	int letterindex;
	png_bytep p;
	/* convert x and y into proper coords */
	x = cx * XGRID;
	y = cy * YGRID+1;

	c = toupper(c);

	/* clear what's there */
	for (yy = y+1; yy <= (y+TEXTSIZEINT);yy++)
		for (xx = x; xx <= (x+TEXTSIZEINT+1); xx++) {
			setpixelrgb(xx,yy,0,0,0);
		}

	switch(c) {
		case '0':
			letterindex = 26;
			break;
		case '1':
			letterindex = 27;
			break;
		case '2':
			letterindex = 28;
			break;
		case '3':
			letterindex = 29;
			break;
		case '4':
			letterindex = 30;
			break;
		case '5':
			letterindex = 31;
			break;
		case '6':
			letterindex = 32;
			break;
		case '7':
			letterindex = 33;
			break;
		case '8':
			letterindex = 34;
			break;
		case '9':
			letterindex = 35;
			break;
		case '-':
			letterindex = 36;
			break;
		case ',':
			letterindex = 37;
			break;
		case '.':
			letterindex = 38;
			break;
		case ':':
			letterindex = 39;
			break;
		case '\'':
			letterindex = 40;
			break;
		case '(':
			letterindex = 41;
			break;
		case ')':
			letterindex = 42;
			break;
		case '!':
			letterindex = 43;
			break;
		case 'A':
			letterindex = 0;
			break;
		case 'B':
			letterindex = 1;
			break;
		case 'C':
			letterindex = 2;
			break;
		case 'D':
			letterindex = 3;
			break;
		case 'E':
			letterindex = 4;
			break;
		case 'F':
			letterindex = 5;
			break;
		case 'G':
			letterindex = 6;
			break;
		case 'H':
			letterindex = 7;
			break;
		case 'I':
			letterindex = 8;
			break;
		case 'J':
			letterindex = 9;
			break;
		case 'K':
			letterindex = 10;
			break;
		case 'L':
			letterindex = 11;
			break;
		case 'M':
			letterindex = 12;
			break;
		case 'N':
			letterindex = 13;
			break;
		case 'O':
			letterindex = 14;
			break;
		case 'P':
			letterindex = 15;
			break;
		case 'Q':
			letterindex = 16;
			break;
		case 'R':
			letterindex = 17;
			break;
		case 'S':
			letterindex = 18;
			break;
		case 'T':
			letterindex = 19;
			break;
		case 'U':
			letterindex = 20;
			break;
		case 'V':
			letterindex = 21;
			break;
		case 'W':
			letterindex = 22;
			break;
		case 'X':
			letterindex = 23;
			break;
		case 'Y':
			letterindex = 24;
			break;
		case 'Z':
			letterindex = 25;
			break;
		default:
			return -1;
	}

	for (yy = 0; yy < TEXTSIZEINT; yy++) {
		p = letter[letterindex].row[yy];
		for (xx = 0; xx < TEXTSIZEINT; xx++) {
			int r,g,b;
			r = *p; p++;
			g = *p; p++;
			b = *p; p++;
			setpixelrgb(x,y,r,g,b);
			x++;
		}
		y++;
		x -= TEXTSIZEINT;
	}
	return 0;
}


void mvaddch(int y, int x, char c) {
	drawchar(y,x,c);
}

void mvaddstr(int y, int x, char *txt) {
	char *p;
	int ind = 0;
	for (p = txt; *p; p++,ind++) {
		drawchar(*p,y,x+ind);
	}
}


/* redraw the whole screen */
int drawscreen(long offset) {
	int y;
	long i;
	char temp[80];
	i = offset;
	for (y = 0;y < nlines;y++) {
		sprintf(temp, "% -25s",game[offset+y].desc);
		mvaddstr(y, TITLEINDENT, temp);
		mvaddstr(y, YEARINDENT, game[offset+y].year);
	}
	/* need to redraw the cursor */
	needupdate = -2;
	newpicture = -1;
	return 0;
}



/* go back to text mode */
void cleanup(void) {
	kill(musicpid, SIGTERM);
	waitpid(musicpid,NULL,0);
	vga_setmode(TEXT);
}

/* figure out how many lines we have to use (was from the ncurses 
 * version) */
long computemaxy(void) {
	long n;
	n = 11;
	return n;
}

/* define colours to use */
void fixpalette(void) {
	gl_setcontext(virtualscreen);

	starcolor[0].r = 0; starcolor[0].g = 252; starcolor[0].b = 248;
	starcolor[1].r = 0; starcolor[1].g = 252; starcolor[1].b = 248;
	starcolor[2].r = 0; starcolor[2].g = 172; starcolor[2].b = 208;
	starcolor[3].r = 0; starcolor[3].g = 96; starcolor[3].b = 168;
	starcolor[4].r = 0; starcolor[4].g = 0; starcolor[4].b = 136;
	starcolor[5].r = 0; starcolor[5].g = 0; starcolor[5].b = 96;

}

/* setup graphics mode, load fonts, etc */
void initgraphics(void) {
	int row;
	int i;
	int x,xx;
	png_bytep p,q;
	png_bytep red,green,blue;
	FILE *f;


	vga_init();
	vga_setmode(GRAPHICSMODE);
	gl_setcontextvga(GRAPHICSMODE);
   	physicalscreen = gl_allocatecontext();
     	gl_getcontext(physicalscreen);

        gl_setcontextvgavirtual(GRAPHICSMODE);
	virtualscreen2 = gl_allocatecontext();
	gl_getcontext(virtualscreen2);

        gl_setcontextvgavirtual(GRAPHICSMODE);
    	virtualscreen = gl_allocatecontext();
       	gl_getcontext(virtualscreen);
	fixpalette();
	
   	for (row = 0; row < MAXSCREENSHOTHEIGHT; row++)
	{
     		//row_pointers[row] = malloc(png_get_rowbytes(png_ptr, info_ptr));
     		row_pointers[row] = malloc(MAXSCREENSHOTWIDTH *3);
 	}

	/* get fonts */
	if (read_png("letters.png")) {
		exit(1);
	}
/*	show_png();
	gl_copyscreen(physicalscreen);
	vga_getch();
	exit(1);*/
	for (i = 0;i < NUMLETTERS; i++) {
		/*int extra;
		int wid;
		switch(i) {
			case 1:
				extra = 1;
				break;
			case 2:
				extra = 1;
				break;
			case 4:
				extra = 0;
				break;
			case 5:
				extra = -1;
				break;
			case 7:
				extra = -4;
				break;
			default:
				extra = 0;
		}*/
		for (row = 0; row < TEXTSIZEINT; row++) {
			/* allocate row */
			letter[i].row[row] = malloc(TEXTSIZEINT*3);
			p = letter[i].row[row];
				
			q = row_pointers[row]+((int)(TEXTSIZE*i)*3 );

			memcpy(p, q, TEXTSIZEINT*3);
		
		}
	}

	/* fix colours */
	for (i = 0;i < NUMLETTERS; i++) {
		for (row = 0; row < TEXTSIZEINT; row++) {
			p = letter[i].row[row];
			for (xx = 0; xx < TEXTSIZEINT; xx++) {
				red = p;
				green = p+1;
				blue = p+2;
				if ((*red <= 20) && (*green <= 20) && (*blue <= 20)) {
					*red = 0;
					*green = 0;
					*blue = 0;
				}
				p += 3;
			}
		}
	}
	
}
	
/* play a random bit of music */
void startmusic(void) {
	char songname[MUSICMAX][15];
	musicpid = fork();
	if (musicpid) {
	} else {
		int i;
		int music = random(4) + 1;
		for (i = 0; i < MUSICMAX; i++ ) {
			sprintf(songname[i], "music%d.mp3",music);
			music++;
			if (music > MUSICMAX) music = 1;
		}
		execlp("mpg123","mpg123","-q",songname[1],songname[2],songname[3],(const char *)NULL);
	}
}

int main(int argc, char **argv) {
	//FILE *f;
	FILE *in;
	long curgame;
	long i;
	char c,c2[2];

	char a,b;
	int end = 0;

	long mamepid = 0;

	int redraw = 0;
	long newoffset = 0;
	int ypos = 0;
	int newypos = 0;
	char temp[90];
	char sstring[90];

	randomize();

	if ((in = fopen(INPUTFILE, "rt")) == NULL) {
		perror(INPUTFILE);
		exit(-1);
	}

	/* read games */
	curgame = 0;
	c = fgetc(in);
	while (!feof(in)) {
		char *p = game[curgame].title;
		int done = 0;
		/* get title */
		p = game[curgame].title;
		for (i = 0; i < 9; i++) {
			if (i <= 7) {/* still in title */
				if (c == ' ') {
					*p = '\0';
					done = -1;
				}
				if (!done) {
					*p = c;
					p++;
				}
				if (i == 7) {
					*p = '\0';
					done = -1;
				}
				
			}
			c = fgetc(in);
		}
		/* get year */
		p = game[curgame].year;
		for (i = 0; i < 4; i++) {
			*p = fgetc(in);
			p++;
		}
		*p = '\0';
		/* skip past maker */
		end = 0;
		a = fgetc(in);
		a = fgetc(in);
		for (;;) {
			b = fgetc(in);
			if (end && (b != ' ')) break;
			else {
				if ((a == ' ') && (b == ' ')) {
					end = 1;
				}
			}
			a = b;
		}

		/* get description */
		p = game[curgame].desc;
		*p = b;
		p++;
		do {
			c = fgetc(in);
			*p = c;
			p++;
		} while ((c != '\n') && (c != '\r'));
		*(p-1) = '\0'; /* replace newline */

		/* max of MAXDESC chars */
		if (((p-2) - game[curgame].desc) >= MAXDESC) {
			game[curgame].desc[MAXDESC] = '\0';
		}
		//printf("%s,%s\n", game[curgame].title, game[curgame].desc);
		curgame++;
		c = fgetc(in);
	}
	//printf("done, %ld games\n",curgame);
	fclose(in);
	/* start music */
	startmusic();
	initgraphics();
	initstars();
	atexit(cleanup);

	nlines = computemaxy();

	gl_setcontext(virtualscreen);

	drawscreen(curoff);
	//gl_copyscreen(physicalscreen);
	//vga_getch();
	//exit(1);

	/*
	vga_setcolor(4);
	linergb(0,0,100,100);
*/
	
	sprintf(temp, "file: % -8s", game[curoff+ypos].title);
	mvaddstr(21,0,temp);
	
	for (;;) {
		int len;
		/* always for now */
		needupdate = -1;
		//mvaddstr(ypos, 0, ">");
		//mvaddstr(ypos, 31, "<");

		removestars();
		
		c = vga_getkey();
		/* move */
		newypos = ypos;
		newoffset = curoff;
		switch (c) {
			case HOME:
				nothingcount = 0;
				imagedir = 1;
				newoffset = 0;
				newypos = 0;
				redraw = -1;
				break;
			case END:
				nothingcount = 0;
				imagedir = 1;
				newoffset = curgame - nlines;
				newypos = (curgame-1) - newoffset;
				redraw = -1;
				break;
			case SLASH:
			case QUESTION:
				nothingcount = 0;
				imagedir = 1;
				mvaddch(nlines,0,c);
				//move(nlines,1);
				c2[1] = '\0';
				c2[0] = '\0';
				strcpy(sstring, "");
				len = 0;
				c2[0] = getchar();
				for (;c2[0] != ENTER;) {
					mvaddch(len,nlines+2,c2[0]);
					strcat(sstring, c2);
					c2[0] = getchar();
				}
			case NEXTSEARCH:
				nothingcount = 0;
				imagedir = 1;
				if ((c == SLASH) || (c == NEXTSEARCH)) {
					int found = 0;
					for(i =(newoffset + newypos) + 1; i < (curgame); i++) {
						if (strstr(game[i].title, sstring) ||
						    strstr(game[i].desc, sstring) ||
						    strstr(game[i].year, sstring)) {
							if (i < (curgame-nlines)) {
								newoffset = i;
								newypos = 0;
							} else {
								newoffset = (curgame - nlines);
								newypos = i-newoffset;
							}
							found = -1;
							redraw = -1;
							break;
						}
					}	
					if (!found) {
						err = -1;
						mvaddstr(nlines, 0, "not found");
					}	
				} else {
					int found = 0;
					for(i =(newoffset + newypos) - 1; i >= 0; i--) {
						if (strstr(game[i].title, sstring) ||
						    strstr(game[i].desc, sstring) ||
						    strstr(game[i].year, sstring)) {
							if (i < (curgame-nlines)) {
								newoffset = i;
								newypos = 0;
							} else {
								newoffset = (curgame - nlines);
								newypos = i-newoffset;
							}
							found = -1;
							redraw = -1;
							break;
						}
					}
					if (!found) {
						err = -1;
						mvaddstr(nlines, 0, "not found");
					}
				}
				break;
					
			case PGUP:
				nothingcount = 0;
				imagedir = 1;
				if (newoffset >= 1) {
					newoffset -= nlines;
					if (newoffset < 0) newoffset = 0;
					redraw = -1;
				} else {
					newypos = 0;
					needupdate = -2;
					newpicture = -1;
				}
				break;
			case PGDOWN:
				nothingcount = 0;
				imagedir = 1;
				if (newoffset < (curgame-nlines)) {
					newoffset += nlines;
					if (newoffset >= (curgame-nlines)) newoffset = curgame-nlines;
					redraw = -1;
				} else {
					newypos = (curgame-1) - newoffset;
					needupdate = -2;
					newpicture = -1;
				}
				break;
			case UP:
				nothingcount = 0;
				imagedir = 1;
				if ((newypos + newoffset) >= 1) {
					if (newypos > 0) {
						newypos--;
						needupdate = -2;
						newpicture = -1;
					} else {
						if (newoffset >= 2) {
							newoffset--;
							redraw = 1;
						}
					}
				}
				break;
			case DOWN:
				nothingcount = 0;
				imagedir = 1;
				if ((newypos + newoffset) <= (curgame-2)) {
					if (newypos < (nlines-1)) {
						newypos++;
						needupdate = -2;
						newpicture = -1;
					} else {
						if (newoffset <= (curgame-2)) {
							newoffset++;
							redraw = 1;
						}
					}
				}
				break;
			case ENTER:
				nothingcount = 0;
				imagedir = 1;
				//def_prog_mode();
				//endwin();
				/* kill music */
				kill(musicpid, SIGTERM);
				waitpid(musicpid,NULL,0);
				/* go to text mode */
				vga_setmode(TEXT);
				mamepid = fork();
				if (mamepid) {
					/* wait for mame to finish */
					waitpid(mamepid,NULL,0);

					/* start music again */
					startmusic();
					/* go back to graphics mode */
					vga_setmode(GRAPHICSMODE);
					fixpalette();

					redraw = -1;
					mvaddstr(nlines,0,"                  ");
					//refresh();
				} else {
					execlp(MAMEPROG,MAMEPROG,game[curoff+ypos].title,NULL);
				}
				break;
			case PNGWRITE:
				nothingcount = 0;
				imagedir = 1;
				drawstars();
				write_png("test.png");
				removestars();
				break;
			case QUIT:
				exit(0);

			default:
		//		sprintf(temp, "unknown char: %d                     ",c);
		//		mvaddstr(nlines, 0, temp);
				nothingcount++;
		}



		movestars();
		
		/* erase cursor */
		if (needupdate) {
			mvaddstr(ypos, TITLEINDENT, game[curoff+ypos].desc);
			mvaddstr(ypos, YEARINDENT, game[curoff+ypos].year);

			linergb(0,ypos*YGRID+1,319,ypos*YGRID+1,0,0,0);
			linergb(0,ypos*YGRID+1,0,ypos*YGRID+YS+3,0,0,0);
			linergb(319,ypos*YGRID+1,319,ypos*YGRID+YS+3,0,0,0);
			linergb(0,ypos*YGRID+YS+4,319,ypos*YGRID+YS+4,0,0,0);
		}


		/* put a delay in - you must stay on the same entry for
		 * a few milliseconds before a new image is loaded, otherwise
		 * we would load them while scrolling, and slow things down */
		if ((imagedir == 0) && (nothingcount >= IMAGEDELAY)) {
			/* start moving back up */
			imagedir = -1;

			/* reset counter */
			nothingcount = 0;

			/* load new image */
			/*
			mvaddstr(ypos, TITLEINDENT, game[curoff+ypos].desc);
			mvaddstr(ypos, YEARINDENT, game[curoff+ypos].year);

			linergb(0,ypos*YGRID,319,ypos*YGRID,0,0,0);
			linergb(0,ypos*YGRID,0,ypos*YGRID+YS+1,0,0,0);
			linergb(319,ypos*YGRID,319,ypos*YGRID+YS+1,0,0,0);
			linergb(0,ypos*YGRID+YS+2,319,ypos*YGRID+YS+2,0,0,0);
	*/
			if (newpicture) {
				/* draw image */
				if (game[newoffset+newypos].title == NULL) {
					read_png("nopic.png");
				} else {
					sprintf(temp, "%s%s.png",ROMPATH,game[newoffset+newypos].title);
					read_png(temp);
				}
				newpicture = 0;
		 	}

		}
		if (imagedir == 1) {
			int donesomething = 0;
			/* move image away, cause we've done something */
			if (SCREENSHOTX+shotxoff < 319) {
				shotxoff+= shotxspeed;
				if (shotxoff > 319) shotxoff = 319;
				donesomething++;
			}
			if (SCREENSHOTY+shotyoff < 239) {
				shotyoff+= shotyspeed;
				if (shotyoff > 239) shotyoff = 239;
				donesomething++;
			}
			if (donesomething != 2)  {
				imagedir = 0;
			}
	
		}
		if (imagedir == -1) {
			int donesomething = 0;


			/* move image up if needed */
			if (shotxoff > 0) {
				shotxoff-= shotxspeed;
				if (shotxoff < 0) shotxoff = 0;
				donesomething = -1;
			}
			if (shotyoff > 0) {
				shotyoff-= shotyspeed;
				if (shotyoff < 0) shotyoff = 0;
				donesomething = -1;
			}

		}

		/* draw screenshot if its moved */
		if ((shotxoff != oldshotxoff) || (shotyoff != oldshotyoff)) {
			/* if it's on the screen */
			if ((SCREENSHOTX+shotxoff < 319) && (SCREENSHOTY+shotyoff < 239)) {
				show_png();
			}
		}
		oldshotxoff = shotxoff;
		oldshotyoff = shotyoff;


		ypos = newypos;
		curoff = newoffset;

		if (redraw) {
			/* redraw screen */
			drawscreen(curoff);
			redraw = 0;
		}

		/* draw cursor */
		if (needupdate) {
		
			/* draw cursor */
			linergb(0,ypos*YGRID+1,319,ypos*YGRID+1,cursorcolor[currentcursor].r,cursorcolor[currentcursor].g,cursorcolor[currentcursor].b);
			linergb(0,ypos*YGRID+1,0,ypos*YGRID+YS+3,cursorcolor[currentcursor].r,cursorcolor[currentcursor].g,cursorcolor[currentcursor].b);
			linergb(319,ypos*YGRID+1,319,ypos*YGRID+YS+3,cursorcolor[currentcursor].r,cursorcolor[currentcursor].g,cursorcolor[currentcursor].b);
			linergb(0,ypos*YGRID+YS+4,319,ypos*YGRID+YS+4,cursorcolor[currentcursor].r,cursorcolor[currentcursor].g,cursorcolor[currentcursor].b);
			needupdate = 0;
		}


		
		drawstars();
		
		if (!err) {
			sprintf(temp, "file: % -8s", game[curoff+ypos].title);
			mvaddstr(21,0,temp);
		} else err = 0;
		
		swapcursorcolors();
		if (brighten) {
			int r,g,b,bx,by;
			gl_copyscreen(virtualscreen2);
			gl_setcontext(virtualscreen2);
			for (by = 0; by < 240; by++) {
				for (bx = 0 ; bx < 320; bx++) {
					getpixelrgb(bx,by,&r,&g,&b);
					r += brighten;
					g += brighten;
					b += brighten;
					if (r > 255) r = 255;
					if (g > 255) g = 255;
					if (b > 255) b = 255;
					setpixelrgb(bx,by,r,g,b);
				}
			}
			vga_waitretrace();
			virttophys();
			gl_setcontext(virtualscreen);
			brighten -= 7;
			if (brighten < 0) brighten = 0;
		} else {
			vga_waitretrace();
			virttophys();
		}



	}

	fclose(in);
	return 0;
}
